#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/netdevice.h>
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <net/ip.h>
#include <net/ipip.h>
#include <linux/if_tunnel.h>
#include <net/mirror.h>

#define DEBUGP(format, args...)

#define dev_safe_put(n)  do { if (n) dev_put(n); (n)=NULL; } while (0)
/* #define dev_up_check(n)  ((n) && ((n)->flags & IFF_UP)) */

static struct net_device* dev_gre0 = NULL;
static struct net_device* dev_wan = NULL;
static char* dev_wan_name = "nas0";
static u_int32_t saddr = 0x101a8c0; /* mirror source ip */
static u_int32_t daddr = 0x201a8c0; /* mirror target ip */
static int  mirror_dir = 3;

/* cmd: insmod mirror dev_wan_name="eth1" saddr=0x101a8c0 daddr=0x201a8c0 mirror_dir=3
 * 		insmod mirror dev_wan_name="ppp0" saddr=0x101a8c0 daddr=0x201a8c0 mirror_dir=3
 */

MODULE_PARM(dev_wan_name, "s");
MODULE_PARM_DESC(dev_wan_name, " name of wan interface: eth1 or ppp0 etc. ");

MODULE_PARM(saddr, "l");
MODULE_PARM_DESC(saddr, " source ip address of our mirror packets ");

MODULE_PARM(daddr, "l");
MODULE_PARM_DESC(daddr, " target ip address of our mirror packets ");

MODULE_PARM(mirror_dir, "i");
MODULE_PARM_DESC(mirror_dir, " the direction of packets to be mirrored ");

static int mirror_send(struct sk_buff* skb)
{
	struct iphdr*  iph;
	int ret = 0;
	
	iph = skb->nh.iph;
	
	DEBUGP("mirror_send(%p):->saddr=%u.%u.%u.%u-->dst=%u.%u.%u.%u (%s)\n",
		skb,NIPQUAD(iph->saddr),NIPQUAD(iph->daddr),skb->dev->name);

	skb->dev = dev_gre0;
	
	ret = dev_queue_xmit(skb);
	
	if(ret != 0){
		DEBUGP("dev_queue_xmit failed ! ! ! ! @..@\n");
	}
	
	return ret;

}

static struct sk_buff* mirror_rcv(struct sk_buff* skb)
{
	/* only handle pkts whose input or output device is wan interface */
	if(skb->dev != dev_wan){
		DEBUGP("mirror_rcv(%p):dev(%s) unexpect\n",skb,skb->dev->name);
		goto rcv_out;
	}
	
	/* only handle ip pkts, skb->protocol have been initialized by driver or L3 layer */
	if(skb->protocol != htons(ETH_P_IP)){
		DEBUGP("mirror_rcv(%p):proto(0x%x) unexpect\n",skb,skb->protocol);
		goto rcv_out;
	}
	
	skb = skb_copy(skb, GFP_ATOMIC);
	if(skb == NULL){
		DEBUGP("skb_copy failed(%p)\n",skb);
		goto rcv_out;
	}
	return skb;
		
rcv_out:	
	return NULL;
}

static int mirror_receive_pkts(struct sk_buff* skb)
{
	DEBUGP("\nmirror_receive_pkts...\n");
	
	skb = mirror_rcv(skb);
	if(skb == NULL)
		return 1;
	
	skb->h.raw = skb->nh.raw = skb->data;
		
	return mirror_send(skb);	
}

static int mirror_xmit_pkts(struct sk_buff* skb)
{
	DEBUGP("\nmirror_xmit_pkts...\n");
	
	skb = mirror_rcv(skb);
	if(skb == NULL)
		return 1;
	
	if(skb_pull(skb, skb->nh.raw - skb->data) == NULL){
		DEBUGP("skb_pull failed(%p)...\n",skb);
		kfree_skb(skb);
		return 1;	
	}
	
	return mirror_send(skb);
}

static void close_gre0(void)
{
	DEBUGP("close_gre0...\n");
	
	if(dev_gre0){
		rtnl_lock();
		dev_close(dev_gre0);
		rtnl_unlock();
		dev_safe_put(dev_gre0);/*dev_get_by_name() hold dev_gre0, release it here*/
	}
}

static int open_gre0(void)
{
	struct net_device* dev;
	struct ip_tunnel *tunnel;
	int ret;
	
	DEBUGP("open_gre0...\n");
	
	dev = dev_get_by_name(DEV_GRE0);
	if(dev == NULL){
		DEBUGP("can not get dev: %s...\n", DEV_GRE0);
		return 1;
	}
	
	/* establish the src ip and target ip for mirror packets*/
	tunnel = (struct ip_tunnel*)dev->priv;	
	tunnel->parms.iph.saddr = saddr;
	tunnel->parms.iph.daddr = daddr;
	
	/* leave this features to be handled by real device */
	dev->features = NETIF_F_SG|NETIF_F_FRAGLIST|NETIF_F_NO_CSUM|NETIF_F_HIGHDMA;
	
	rtnl_lock();
	ret = dev_open(dev);
	rtnl_unlock();
	
	if(ret){
		DEBUGP("dev_open(%s) failed...\n",dev->name);
		dev_put(dev);
		return 1;
	}
	
	dev_safe_put(dev_gre0);
	dev_gre0 = dev;

	return 0;
}

static inline 
void _mirror_hook_down(void)
{
	mirror_receive_pkts_hook = NULL;
	mirror_xmit_pkts_hook = NULL;
}

static inline
void _mirror_hook_up(void)
{
	switch(mirror_dir){
		case DIR_TX:
			mirror_xmit_pkts_hook = mirror_xmit_pkts;
			break;
		case DIR_RX:
			mirror_receive_pkts_hook = mirror_receive_pkts;
			break;
		case DIR_BOTH:
			mirror_receive_pkts_hook = mirror_receive_pkts;
			mirror_xmit_pkts_hook = mirror_xmit_pkts;
			break;
		case DIR_OFF:
			break;
		default:
			break;
	}	
}

/* it's necessary to disable bottom halves here, for hooks work in bottom halves env. */
static void mirror_disable(void)
{
	DEBUGP("\nmirror_disable...\n");
	
	write_lock_bh(&mirror_rw_lock);
	_mirror_hook_down();
	dev_safe_put(dev_wan);
	write_unlock_bh(&mirror_rw_lock);
}

static void mirror_enable(void)
{
	struct net_device* dev;
	
	DEBUGP("\nmirror_enable...\n");	
	
	dev = dev_get_by_name(dev_wan_name);
	if(dev == NULL){
		DEBUGP("can not get dev: %s\n",dev_wan_name);
		return;
	}
	
	/* it's necessary to get lock before check dev state, I don't wana NETDEV_DOWN 
	 * to break in after check to run mirror_disable (it wins the race)
	 */
	write_lock_bh(&mirror_rw_lock);
	if(netif_running(dev) && netif_running(dev_gre0)){		
		/* release dev_wan if it have been hold (mirror have been enabled), so 
		 * it will be OK to call mirror_enable() at bath cases (enable or disable) 
		 */
		dev_safe_put(dev_wan);
		dev_wan = dev;
		_mirror_hook_up();
		
	}
	else{
		dev_put(dev);
	}
	write_unlock_bh(&mirror_rw_lock);	
}

/* a little rough @_@ */
static int mirror_netdev_event(struct notifier_block* self, 
		unsigned long event, void* ptr)
{
	struct net_device* dev = (struct net_device*)ptr;
	
	switch(event){	
		case NETDEV_DOWN:		
			DEBUGP("mirror: NETDEV_DOWN(%s)\n", dev->name);
			if(strcmp(dev->name,dev_wan_name)==0
				|| strcmp(dev->name,DEV_GRE0)==0){
				mirror_disable();
			}
			break;
		case NETDEV_UP:
			DEBUGP("mirror: NETDEV_UP(%s)\n", dev->name);
			if(strcmp(dev->name,dev_wan_name)==0
				|| strcmp(dev->name,DEV_GRE0)==0){
				mirror_enable();
			}
			break;
		default:
			break;
	}
	return NOTIFY_DONE;
}

static struct notifier_block mirror_netdev_notifier = {
	.notifier_call = mirror_netdev_event,
};

static int __init mirror_init(void)
{
	printk("mirror_init: module mirror startup...\n");
	
	/* just leave dev_gre0 out of critical region for simplify */
	if(open_gre0())
		return 1;
	
	register_netdevice_notifier(&mirror_netdev_notifier);
	
	mirror_enable();
	
	return 0;
}

static void __exit mirror_fini(void)
{
	printk("mirror_fini: module mirror exit...\n");
	
	unregister_netdevice_notifier(&mirror_netdev_notifier);
	mirror_disable();
	close_gre0();
}

module_init(mirror_init);
module_exit(mirror_fini);
